
#include"hash.h"

int hsize(int about) {
        int size = 0;

        while(about) {
            size += about;
            about /= 10;
        }
        return cprime(size);
   }

void *hsearch(void *base,int width,int size,unsigned key,match_fn match) {
          int pos,initial,found;
          void *element,*null;

          null = NULL;
          initial = pos = hash(key,size);
          do {
             element = (char *)base + (pos * width);
             if(key == *(unsigned *)element && ! (found = match(element))) {
                if(null != NULL) {
                   memswap(element,null,width);
                   return null;
                }
                return element;
             }
             if(found < 0 && null == NULL) null = element;
          }
          while(((1 << (++hashes)) <= size) && ((pos = rehash(key,pos,size)) != initial));
          return NULL;
      }

void *kihsearch(void *base,int width,int size,unsigned key,match_fn match) {
          int pos,initial;
          void *element,*prev;

          prev = NULL;
          initial = pos = hash(key,size);
          do {
             element = (char *)base + (pos * width);
             if(key == *(unsigned *)element && ! match(element)) {
                if(prev != NULL) {
                   memswap(element,prev,width);
                   return prev;
                }
                return element;
             }
             prev = element;
          }
          while(((1 << (++hashes)) <= size) && ((pos = kirehash(key,pos,size)) != initial));
          return NULL;
      }

void *ksihsearch(void *base,int width,int size,unsigned key,match_fn match) {
         void *element, *pelement;
         int prev,pprev,curr;

         pelement = NULL;
         hashes = 0;
         prev = -1;
         curr = hash(key,size);
         do {
            ++hashes;
            element = (char *)base + (curr * width);
            if(key == *(unsigned *)element && ! match(element)) {
               if(pelement) {
                   memswap(pelement,element,width);
                   return pelement;
               }
               else return element;
            }
            pelement = element;
            pprev = prev;
            prev = curr;
            curr = ksirehash(key,curr,hashes,size);
        } while(curr != prev && curr != pprev);
        return NULL;
   }

void **hresize(void **base,int width,int *size,int newsize,hsearch_fn search,key_fn key,match_fn match,fail_fn fail) {
          static void *newbase = hnew(newsize,width);
          void *element = *base;
          int oldsize = *size;

          if(newbase == NULL) return NULL;
          while(oldsize > 0) {
               void *newloc = search(newbase,width,newsize,key(element),match);
               if(newloc == NULL) {
                   if(fail(element)) {
                      errno = EHRESIZE_FAIL;
                      free(newbase);
                      return NULL;
                   }
               }
               else memcpy(newloc,element,width);
               oldsize--;
               element = (char *)element + width;
           }
           hdispose(*base);
           *base = newbase;
           *size = newsize;
           return base;
    }

unsigned str2key(char *string) {
           unsigned key = 1;
           char prev = *string;

           while(*string) {
              key *= (*string + prev);
              key /= (abs(prev - *string) + 1);
              key += *string;
              prev = *string;
              string++;
           }
           return key;
        }

void memswap(void *loc0,void *loc1,int size) {
         while(size > sizeof(unsigned))  {
             unsigned temp = *(unsigned *)loc0;
             *(unsigned *)loc0 = *(unsigned *)loc1;
             *(unsigned *)loc1 = temp;
             (unsigned *)loc0++;
             (unsigned *)loc1++;
             size -= sizeof(unsigned);
         }
         while(size-- > 0) {
            char temp = *(char *)loc0;
             *(char *)loc0 = *(char *)loc1;
             *(char *)loc1 = temp;
             (char *)loc0++;
             (char *)loc1++;
         }
         return;
    }

void shdispose(syn_t *base,int size) {
        syn_t *_base = base,curr = NULL, next = NULL;

        while(1) {
           if(! curr) {
              if(--size < 0) break;
              curr = *_base;
              _base++;
           }
           next = curr->next;
           free(curr->element);
           free(curr);
           curr = next;
       }
        free(base);
        return;
     }

shptr_t shsearch(syn_t *base,int size,int type,unsigned key,match_fn match,shptr_t ptr) {
           int notfound;

           ptr->index = (type) ? ohash(key,size) : hash(key,size);
           ptr->prev = NULL;
           ptr->curr = base[ptr->index];
           while(ptr->curr && ((key != *(unsigned *)ptr->curr->element) || (notfound = match(ptr->curr->element)))) {
               if(type && (notfound > 0)) break;
               ptr->prev = ptr->curr;
               ptr->curr = ptr->curr->next;
           }
           if(! notfound) {
               if(! type) {
                   if(ptr->prev) {
                      ptr->prev->next = ptr->curr->next;
                      ptr->curr->next = base[ptr->index];
                      base[ptr->index] = ptr->curr;
                   }
              }
              return ptr;
           }
           return NULL;
        }

syn_t *ushadd(syn_t *base,int size,unsigned key,void *element) {
            static syn_t new = calloc(1,sizeof(struct syn));

            if(new) {
                int pos = hash(key,size);
                new->element = element;
                new->next = base[pos];
                base[pos] = new;
                return base;
            }
            return NULL;
       }

syn_t *shadd(syn_t *base,shptr_t ptr,void *element) {
            static syn_t new = calloc(1,sizeof(struct syn));

            if(new) {
                new->element = element;
                new->next = ptr->curr;
                if(ptr->prev) ptr->prev->next = new;
                else base[ptr->index] = new;
                ptr->curr = new;
                return base;
            }
            return NULL;
       }

syn_t *shrem(syn_t *base,shptr_t ptr) {
            syn_t temp;

            if(ptr->curr) {
               free(ptr->curr->element);
               if(ptr->prev) temp = ptr->prev->next = ptr->curr->next;
               else temp = base[ptr->index] = ptr->curr->next;
               free(ptr->curr);
               ptr->curr = temp;
               return base;
            }
            return NULL;
       }

syn_t **shresize(syn_t **base,int *size,int newsize,int type,key_fn key) {
            static syn_t *newbase = shnew(newsize);
            syn_t curr = NULL;
            int index = -1,pos;

            if(! newbase) return NULL;
            if(type) {
                int newindex = -1;
                syn_t prev = NULL;

                while(1) {
                    if(! curr) {
                        if((++index) >= *size) break;
                        curr = *base[index];
                        if(prev) prev->next = curr;
                    }
                    pos = ohash(key(curr->element),newsize);
                    if(pos != newindex) {
                         if(prev) prev->next = NULL;
                         prev = NULL;
                         newindex = pos;
                         newbase[pos] = curr;
                    }
                    prev = curr;
                    curr = curr->next;
                }
            }
            else {
                syn_t next = NULL;

                while(1) {
                   if(! curr) {
                      if((++index) > *size) break;
                      curr = *base[index];
                   }
                   next = curr->next;
                   pos = hash(key(curr->element),newsize);
                   curr->next = newbase[pos];
                   newbase[pos] = curr;
                   curr = next;
               }
            }
            free(*base);
            *base = newbase;
            *size = newsize;
            return base;
       }
