#ifndef HASH_H
#define HASH_H

#include<stdlib.h>
#include<string.h>
#include<ctype.h>
#include<limits.h>
#include<errno.h>
#include"prime.h"

typedef unsigned (* key_fn)(void *element);
typedef int (* match_fn)(void *element);
typedef void *(* hsearch_fn)(void *base,int width,int size,unsigned key,match_fn match);
typedef int (* fail_fn)(void *element);

#define hash(key,size) ((key) % (size))
#define rehash(key,prev,size) hash((key)*((prev) + 1),size)
#define kirehash(key,prev,size) hash((key) + (prev),size)
#define ksirehash(key,prev,time,size) hash((key) + (prev) + (time),size)

int hsize(int about);
#define hnew(size,width) calloc(size,width)
#define hdispose(base) free(base)

int hashes;
void *hsearch(void *base,int width,int size,unsigned key,match_fn match);
void *kihsearch(void *base,int width,int size,unsigned key,match_fn match);
void *ksihsearch(void *base,int width,int size,unsigned key,match_fn match);

#define resize(size,count,ops) ((size) * ((count) + (ops)) / (ops))

#define hnewsize(size,count,ops) hsize(resize(size,count,ops))
void **hresize(void **base,int width,int *size,int newsize,hsearch_fn search,key_fn key,match_fn match,fail_fn fail);
#define EHRESIZE_FAIL (-1923)

unsigned str2key(char *string);
void memswap(void *loc0,void *loc1,int size);

struct syn {
        void *element;
        struct syn *next;
};
typedef struct syn *syn_t;

struct shptr {
       int index;
       struct syn *prev, *curr;
};
typedef struct shptr *shptr_t;

#define ohash(key,size)  (((key) * (size)) / UINT_MAX)

#define HTABLE (0)
#define OHTABLE (!0)

#define shsize(about) cprime(about)
#define shnew(size) ((syn_t *)calloc(size,sizeof(syn_t)))
void shdispose(syn_t *base,int size);

syn_t *ushadd(syn_t *base,int size,unsigned key,void *element);
shptr_t shsearch(syn_t *base,int size,int type,unsigned key,match_fn match,shptr_t ptr);
syn_t *shadd(syn_t *base,shptr_t ptr,void *element);
syn_t *shrem(syn_t *base,shptr_t ptr);

#define shnewsize(size,count,ops) shsize(resize(size,count,ops))
syn_t **shresize(syn_t **base,int *size,int newsize,int type,key_fn key);

#endif
